/**
 * conjecture_wormald.c
 * 
 * Copyright (c) 2016-2017 Ghent University
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/times.h>
#include "nauty.h"
#include "util.h"

#define VERSION "1.0 - May 18 2017"

//Test for the strong version of Wormald's conjecture:
//Let G be a cubic graph with |E(G)| = 0 mod 2. Then there exists a 
//2-edge-colouring of E(G) such that the two monochromomatic subgraphs 
//are isomorphic linear forests AND each path has length at least 2.

/*************************Defines and global variables*************************/

//#define POPC(x) (__builtin_popcountll(x)) 
#define BIT(i) (1ULL << (i))


int valid_colouring_found = 0;

int num_marked_edges = 0;
//unsigned long long int marked_edges_bitvector = 0;

//The neighbourhood of each vertex (the vertex itself is not included in the neighbourhood)
unsigned long long int neighbourhood[MAXN];

EDGE edgelist[3*MAXN/2];
int num_edges = 0;

int edgelabel[3*MAXN/2][3*MAXN/2];


/* Some marks */
//Edge marks!
unsigned int marks[3*MAXN/2];

#define MAXVAL INT_MAX - 1
static int markvalue = MAXVAL;
#define RESETMARKS {int mki; if ((markvalue += 1) > MAXVAL) \
      { markvalue = 1; for (mki=0;mki<(3*MAXN/2);++mki) marks[mki]=0;}}
#define MARK(v) marks[v] = markvalue
#define UNMARK(v) marks[v] = markvalue - 1
#define ISMARKED(v) (marks[v] == markvalue)



unsigned int marks2[MAXN];
static int markvalue2 = MAXVAL;
#define RESETMARKS2 {int mki; if ((markvalue2 += 1) > MAXVAL) \
      { markvalue2 = 1; for (mki=0;mki<MAXN;++mki) marks2[mki]=0;}}
#define MARK2(v) marks2[v] = markvalue2
#define UNMARK2(v) marks2[v] = markvalue2 - 1
#define ISMARKED2(v) (marks2[v] == markvalue2)


int number_of_components = 0;
int number_of_vertices_marked = 0;
int number_of_vertices_in_component = 0;

int number_of_edges_in_component = 0;


GRAPH marked_graph;
ADJACENCY marked_adj;

GRAPH unmarked_graph;
ADJACENCY unmarked_adj;

/******************************************************************************/

//For nauty
#define WORKSIZE (50 * MAXM)

int lab[MAXN], ptn[MAXN], orbits[MAXN];
static DEFAULTOPTIONS_GRAPH(options);
statsblk stats;
setword workspace[WORKSIZE];

//graph nautyg[MAXN*MAXM];

/******************************************************************************/

int num_2edge_colourings_generated = 0;

//If = 1 it tests the strong version of the conjecture, else the original version
int test_strong_conjecture = 0;


int test_thomassen_conjecture = 0;

/*********************************Methods**************************************/

//Separate methods is not faster!

/*
//Only testing vertices with label 0..maxnv-1
//If contains two or more marked neighbours, then there is a monochromatic component of size at least three
//If reverse = 1, we search for vertices which are coloured white (i.e. unmarked vertices)
int each_marked_vertex_contains_at_most_one_marked_nbr(unsigned long long int vertices_bitvector, 
        int reverse, int maxnv) {
    
    unsigned long long int valid_vertices_mask = BIT(maxnv) - 1;
    vertices_bitvector &= valid_vertices_mask;
    
    //Verified that going through all neighbours without POPC is a lot slower!
    int i;
    for(i = 0; i < maxnv; i++)
        if((!reverse && ISMARKED(i)) || (reverse && !ISMARKED(i))) {
            int num_marked_nbrs = POPC(neighbourhood[i] & vertices_bitvector);
            if(num_marked_nbrs > 1)
                return 0;
        }
    
    return 1;
}
*/

/******************************************************************************/

/**
 * Prints the given nauty graph.
 */
static void
printgraph_nauty(graph *g, int current_number_of_vertices) {
    int i;
    fprintf(stderr, "Printing graph nauty:\n");
    for(i = 0; i < current_number_of_vertices; i++) {
        fprintf(stderr, "%d :", i);
        set *gv = GRAPHROW1(g, i, MAXM);
        int neighbour = -1;
        while((neighbour = nextelement(gv, MAXM, neighbour)) >= 0) {
            fprintf(stderr, " %d", neighbour);
        }
        fprintf(stderr, "\n");
    }
}

/******************************************************************************/

/*
 * Converts g to a graph in the nauty format and stores the resulting graph
 * in nauty_graph.
 */
static void
copy_nauty_graph_marked_edges(graph *nauty_graph, int use_marked) {
    int i;
    for(i = 0; i < nv; i++)
        nauty_graph[i] = 0;
    
    for(i = 0; i < num_edges; i++)
        if(ISMARKED(i) == use_marked) {
            ADDELEMENT1(&nauty_graph[edgelist[i][0]], edgelist[i][1]);
            ADDELEMENT1(&nauty_graph[edgelist[i][1]], edgelist[i][0]);
        }
    
/*
    int i, j;
    for(i = 0; i < num_vertices; i++) {
        nauty_graph[i] = 0;
        if(ISMARKED(i) == use_marked)
            for(j = 0; j < adj[i]; j++)
                if(ISMARKED(g[i][j]) == use_marked)
                    ADDELEMENT1(&nauty_graph[i], g[i][j]);
    }
*/

}

/******************************************************************************/

void compute_canon_form_graph_induced_by_marked_edges(graph *nautyg_canon, int use_marked) {
    options.defaultptn = TRUE;
    options.getcanon = TRUE;      
    
    graph nautyg_temp[MAXN*MAXM];
    copy_nauty_graph_marked_edges(nautyg_temp, use_marked);
    
    //fprintf(stderr, "graph induced by marked edges:\n");
    //printgraph_nauty(nautyg_temp, nv);
   
    nauty(nautyg_temp, lab, ptn, NULL, orbits, &options, &stats, workspace, WORKSIZE, 
            MAXM, nv, nautyg_canon);    
    
}

/******************************************************************************/

int contains_vertex_with_more_than_two_unmarked_nbrs() {
/*
    int res_new = contains_vertex_with_more_than_two_unmarked_nbrs_new();
    
    int res_old = contains_vertex_with_more_than_two_marked_nbrs_old(0);
    
    if(res_new != res_old) {
        fprintf(stderr, "Error: different results!\n");
        exit(1);
    }
    
    return res_new;    
*/
    
    int i;
    for(i = 0; i < nv; i++)
        if(unmarked_adj[i] > 2)
            return 1;
    
    return 0;
}

/******************************************************************************/

int contains_vertex_without_marked_nbrs(int use_marked) {
/*
    int i, j;
    for(i = 0; i < nv; i++) {
        int num_marked_nbrs = 0;
        for(j = 0; j < adj[i]; j++)
            if(ISMARKED(edgelabel[i][current_graph[i][j]]) == use_marked)
                num_marked_nbrs++;
        
        //Never happens
        if(num_marked_nbrs == 0) {
            return 1;
        }
    }
    
    return 0;
*/
    
    int i;
    if(use_marked) {
        for(i = 0; i < nv; i++)
            if(marked_adj[i] == 0)
                return 1;
    } else {
        for(i = 0; i < nv; i++)
            if(unmarked_adj[i] == 0)
                return 1;        
    }
    
    return 0;
}

/******************************************************************************/

/**
 * DFS algorithm which recursively marks all vertices which can be reached from
 * current_vertex.
 * Complexity: O(|V|+|E|)
 */
static void mark_dfs_marked(unsigned char current_vertex) {
    MARK2(current_vertex);
    number_of_vertices_marked++;
    number_of_vertices_in_component++;
    number_of_edges_in_component += marked_adj[current_vertex];
    
    //fprintf(stderr, "marking %d\n", current_vertex);
    
    int i;
    for(i = 0; i < marked_adj[current_vertex]; i++)
        if(!ISMARKED2(marked_graph[current_vertex][i])) {
            mark_dfs_marked(marked_graph[current_vertex][i]);
        }
    
}

/******************************************************************************/

static void mark_dfs_unmarked(unsigned char current_vertex) {
    MARK2(current_vertex);
    number_of_vertices_marked++;
    number_of_vertices_in_component++;
    number_of_edges_in_component += unmarked_adj[current_vertex];
    
    //fprintf(stderr, "marking %d\n", current_vertex);
    
    int i;
    for(i = 0; i < unmarked_adj[current_vertex]; i++)
        if(!ISMARKED2(unmarked_graph[current_vertex][i])) {
            mark_dfs_unmarked(unmarked_graph[current_vertex][i]);
        }
    
}

/******************************************************************************/

int monochromatic_component_graphs_are_acyclic_and_each_component_has_at_least_3_vertices(int use_marked) {
    
    number_of_vertices_marked = 0;
    number_of_components = 0;
    
    int smallest_component_nv = MAXN;
    
    //It is no problem if there would be isolated vertices...
    RESETMARKS2;
    int i;
    for(i = 0; i < nv; i++)
        if(!ISMARKED2(i)) {
            number_of_vertices_in_component = 0;
            number_of_edges_in_component = 0;
            
            //fprintf(stderr, "start from vertex %d\n", i);
            
            if(use_marked)
                mark_dfs_marked(i);
            else
                mark_dfs_unmarked(i);
            number_of_components++;
            
            //Counted very edge twice!
            number_of_edges_in_component /= 2;
            
/*
            int ne_in_comp_old = determine_number_of_edges_in_component(current_component_bitvector, use_marked);
            
            if(ne_in_comp_old != number_of_edges_in_component) {
                fprintf(stderr, "Error: wrong ne! %d vs %d\n", ne_in_comp_old, number_of_edges_in_component);
                exit(1);
            }
*/
            
            //fprintf(stderr, "number_of_vertices_in_component: %d, number_of_edges_in_component: %d\n",
            //        number_of_vertices_in_component, number_of_edges_in_component);                   
            
            //Is not acyclic
            if(number_of_edges_in_component != number_of_vertices_in_component - 1) {
                return 0;
            }

            //if(test_thomassen_conjecture && number_of_vertices_in_component > 6) {
            //Below is a hack which requires that the paths must have length at most 4 instead of 5 
            //(so at most 5 vertices)
            if(test_thomassen_conjecture && number_of_vertices_in_component > 5) {
/*
                printgraph(current_graph, adj);

                int k;
                for(k = 0; k < num_edges; k++)
                    if(ISMARKED(k) == use_marked)
                        fprintf(stderr, "edge %d %d\n", edgelist[k][0], edgelist[k][1]);

                fprintf(stderr, "use_marked ? %d\n", use_marked);

                fprintf(stderr, "pathlength not ok!\n");
                fprintf(stderr, "vertex: %d\n", i);
*/

                //Path with more than 6 vertices (more than 5 edges)
                return 0;
            }
            
            if(number_of_vertices_in_component < smallest_component_nv) {
                smallest_component_nv = number_of_vertices_in_component;
                
                //Each path has length at least 2 (i.e. at least 3 vertices)
                if(test_strong_conjecture && smallest_component_nv < 3) {
/*
                    printgraph(current_graph, adj);

                    int k;
                    for(k = 0; k < num_edges; k++)
                        if(ISMARKED(k) == use_marked)
                            fprintf(stderr, "edge %d %d\n", edgelist[k][0], edgelist[k][1]);

                    fprintf(stderr, "use_marked ? %d\n", use_marked);                    
                    
                    fprintf(stderr, "pathlength not ok!\n");
                    fprintf(stderr, "smallest_component_nv: %d\n", smallest_component_nv);
                    fprintf(stderr, "vertex: %d\n", i);
*/

                    return 0;
                }
            }
        }
    
    if(number_of_vertices_marked != nv) {
        fprintf(stderr, "Error: invalid nv marked!\n");
        exit(1);
    }
    
    return 1;
}

/******************************************************************************/

int graphs_induced_by_monochromatic_components_are_isomorphic() {
    
/*
    fprintf(stderr, "org graph:\n");
    printgraph(current_graph, adj);
   
    fprintf(stderr, "marked vertices:\n");
    int i;
    for(i = 0; i < nv; i++)
        if(ISMARKED(i))
            fprintf(stderr, "%d\n", i);    
*/
    
    

    //We only need to consider paths!
    
    //This was already tested before!
/*
    if(contains_vertex_with_more_than_two_marked_nbrs(1)) {
        fprintf(stderr, "Error: contains vertex with more than 2 marked nbrs (is already tested before!)\n");
        exit(1);
    }
*/
    
    //if(contains_vertex_with_more_than_two_marked_nbrs(1)
    //        || contains_vertex_with_more_than_two_marked_nbrs(0)) {
    //if(contains_vertex_with_more_than_two_marked_nbrs(0)) {
    if(contains_vertex_with_more_than_two_unmarked_nbrs()) {
        return 0;
    }    

    if(test_strong_conjecture) {
        //Isolated vertices are forbidden since paths must have at least 2 edges
        if(contains_vertex_without_marked_nbrs(1)
                || contains_vertex_without_marked_nbrs(0)) {
            //fprintf(stderr, "Isolated vertex found?!\n"); //Never happens?
            return 0;
        }
    }
    
    //Testing if the marked components are acyclic sooner using fast method is slower!
    //So therefore only testing them now...
    
    //First test if the induced graphs are acyclic (i.e. forests)
    if(!monochromatic_component_graphs_are_acyclic_and_each_component_has_at_least_3_vertices(1)
            || !monochromatic_component_graphs_are_acyclic_and_each_component_has_at_least_3_vertices(0))
        return 0;    
    
    
    //Hack no isomorphism testing!
    if(test_thomassen_conjecture) //No iso testing for thomassen's conjecture!
        return 1;
    
    //Note that the marked graphs can have isolated edges (and this is exactly the definition)
    //Note that for the strong version of Wormald's conjecture isolated vertices are forbidden
    //(since each path must have length at least 2, i.e. 2 edges (= 3 vertices))
    
    graph nautyg_marked_canon[MAXN*MAXM];
    compute_canon_form_graph_induced_by_marked_edges(nautyg_marked_canon, 1);
    
    //fprintf(stderr, "canon form:\n");
    //printgraph_nauty(nautyg_marked_canon, nv);            

    graph nautyg_unmarked_canon[MAXN*MAXM];
    compute_canon_form_graph_induced_by_marked_edges(nautyg_unmarked_canon, 0);    
    
    //fprintf(stderr, "canon form unmarked:\n");
    //printgraph_nauty(nautyg_unmarked_canon, nv);            
    
    if(memcmp(nautyg_marked_canon, nautyg_unmarked_canon, nv * sizeof(graph)) == 0) {
        //fprintf(stderr, "Found isomorphic graphs!\n");

/*
        printgraph(current_graph, adj);

        fprintf(stderr, "colouring found!\n");
        
        fprintf(stderr, "Red edges:\n");
        
        int k;
        for(k = 0; k < num_edges; k++)
            if(ISMARKED(k) == 1)
                fprintf(stderr, "{%d %d} ", edgelist[k][0], edgelist[k][1]);
                //fprintf(stderr, "edge %d %d\n", edgelist[k][0], edgelist[k][1]);
        fprintf(stderr, "\n");

        exit(1);
*/
        
        return 1;
    }
    
    return 0;
}

/******************************************************************************/

/*
int contains_isolated_unmarked_edge_old(int current_edge) {
    
    int i, j;
    for(j = 0; j < 2; j++)
        for(i = 0; i < adj[edgelist[current_edge][j]]; i++) {
            if(is_isolated_unmarked_edge(edgelabel[edgelist[current_edge][j]][current_graph[edgelist[current_edge][j]][i]], 0))
                return 1;
        }

    return 0;
}
*/

int contains_isolated_unmarked_edge(int current_edge) {
    
/*
    if(contains_isolated_unmarked_edge_new(current_edge) != contains_isolated_unmarked_edge_old(current_edge)) {
        fprintf(stderr, "Error: wrong result!\n");
        exit(1);
    }
    
    return contains_isolated_unmarked_edge_new(current_edge);
*/
    
    int i;
    for(i = 0; i < 2; i++) {
        if(unmarked_adj[edgelist[current_edge][i]] == 1
                && unmarked_adj[unmarked_graph[edgelist[current_edge][i]][0]] == 1)
            return 1;
    }
    
    return 0;    
}

/******************************************************************************/

//Adj lists are not sorted, but is no problem...
static void
remove_edge_from_neighbourhood(GRAPH given_graph, unsigned char given_adj[],
        unsigned char vertex, unsigned char removed_vertex) {
    int i;
    for(i = 0; i < given_adj[vertex]; i++)
        if(given_graph[vertex][i] == removed_vertex)
            break;
    if(i < given_adj[vertex]) {
        i++;
        while(i < given_adj[vertex]) {
            given_graph[vertex][i-1] = given_graph[vertex][i];
            i++;
        }
        given_adj[vertex]--;
    } else {
        fprintf(stderr, "Error: vertex which is going to be removed was not a neighbour!\n");
        fprintf(stderr, "%d %d\n", vertex, removed_vertex);
        exit(1);
    }    
}

/******************************************************************************/

static void 
add_edge_given_graph(GRAPH given_graph, unsigned char given_adj[], unsigned char v, unsigned char w) {
    given_graph[v][given_adj[v]] = w;
    given_graph[w][given_adj[w]] = v;
    given_adj[v]++;
    given_adj[w]++;
}

/******************************************************************************/

int marked_component_contains_cycle_or_more_than_5_vertices(unsigned char vertex) {
    RESETMARKS2
    number_of_vertices_in_component = 0;
    number_of_edges_in_component = 0;

    mark_dfs_marked(vertex);


    //Counted very edge twice!
    number_of_edges_in_component /= 2;


    //Is not acyclic
    if(number_of_edges_in_component != number_of_vertices_in_component - 1) {
        return 1;
    }

    //if(test_thomassen_conjecture && number_of_vertices_in_component > 6) {
    //Below is a hack which requires that the paths must have length at most 4 instead of 5 
    //(so at most 5 vertices)
    if(test_thomassen_conjecture && number_of_vertices_in_component > 5) {
        return 1;
    }

    return 0;
}

/******************************************************************************/

void colour_edges_in_all_possible_ways_recursion(int next_edge) {
    
    if(valid_colouring_found)
        return;
    
    if(num_marked_edges > num_edges / 2) {
        fprintf(stderr, "Error: num marked num_edges is too high!\n");
        exit(1);
    }
    
    if(num_marked_edges == num_edges / 2) {
    //if(num_marked_edges <= num_edges / 2) {
        //Test if it is valid colouring
        
/*
        if(POPC(marked_edges_bitvector) != nv / 2) {
            fprintf(stderr, "Error: not all vertices in bitvector!\n");
            exit(1);
        }
*/
        
        num_2edge_colourings_generated++;
        
        //Test if this is a valid bisection
        //Found a valid bisection, so we can stop!
        if(graphs_induced_by_monochromatic_components_are_isomorphic())
            valid_colouring_found = 1;
        
        return;
    } 
    
/*
    else {
        int num_available_vertices = nv - next_vertex;
        if(num_marked_edges + num_available_vertices < nv / 2) {
            fprintf(stderr, "Never happens...\n");
            return; //Not enough available vertices left
        }
    }
*/
    
    if(next_edge >= num_edges) {
        fprintf(stderr, "Error: should never happen\n");
        exit(1);
    }
    
    
    //TODO: ook beter niet in random volgorde testen?

    int is_still_valid_colouring_reverse = 1;
    int i;
    for(i = next_edge; i < num_edges && is_still_valid_colouring_reverse; i++) {
        MARK(i);
        num_marked_edges++;
        //marked_edges_bitvector |= BIT(i);        
        
        //Helps a bit
        int num_available_edges = num_edges - (i + 1);
        if(num_marked_edges + num_available_edges < num_edges / 2)
            is_still_valid_colouring_reverse = 0;

        if(is_still_valid_colouring_reverse) {
            int can_continue = 1;

            //Check if no vertex with more than 2 marked nbrs (has to be endpoint of this edge)

            //can_continue = !contains_vertex_with_more_than_two_marked_nbrs(1);
            //can_continue = !given_vertex_has_more_than_two_marked_nbrs(edgelist[i][0], 1)
            //        && !given_vertex_has_more_than_two_marked_nbrs(edgelist[i][1], 1);
            //can_continue = marked_adj[edgelist[i][0]] <= 2 && marked_adj[edgelist[i][1]] <= 2;
            
            //Adding an edge will increase marked_adj by one
            can_continue = marked_adj[edgelist[i][0]] <= 1 && marked_adj[edgelist[i][1]] <= 1;

            if(can_continue) {
                
                //It is best to postpone the actual removed of the edges in the unmarked graph!
                add_edge_given_graph(marked_graph, marked_adj, edgelist[i][0], edgelist[i][1]);
                remove_edge_from_neighbourhood(unmarked_graph, unmarked_adj, edgelist[i][0], edgelist[i][1]);
                remove_edge_from_neighbourhood(unmarked_graph, unmarked_adj, edgelist[i][1], edgelist[i][0]);


                //Testing if component of last edge contains a cycle is much slower, so don't test this here!

                if(can_continue && test_strong_conjecture) {
                    //Test if there is an unmarked component with less than 3 vertices...
                    /*
                                    if(contains_vertex_without_marked_nbrs(0)) {
                                        fprintf(stderr, "ok...\n");
                                        //Never happens???
                                    }
                     */
                    //Doesn't really help much to postpone adding/removing edges still this test
                    if(contains_isolated_unmarked_edge(i)) 
                        can_continue = 0;
                }

                if(can_continue && test_thomassen_conjecture) {
                    //Test if component where last marked edge is part of contains
                    //more than 5 vertices or is part of a cycle!

                    //Only testing this for thomassen conjecture since just testing if marked component
                    //contains a cycle is slower (e.g. for Wormald's conjecture)

                    if(marked_component_contains_cycle_or_more_than_5_vertices(edgelist[i][0]))
                        can_continue = 0;

                }

                if(can_continue)
                    colour_edges_in_all_possible_ways_recursion(i + 1);

                //remove_edge_from_neighbourhood(marked_graph, marked_adj, edgelist[i][0], edgelist[i][1]);
                //remove_edge_from_neighbourhood(marked_graph, marked_adj, edgelist[i][1], edgelist[i][0]);
                //Expensive while we know the nbr is the last one, so use:
                marked_adj[edgelist[i][0]]--;
                marked_adj[edgelist[i][1]]--;

                add_edge_given_graph(unmarked_graph, unmarked_adj, edgelist[i][0], edgelist[i][1]);
                
        
                if(valid_colouring_found)
                    return;                        
                
            }

        }
        
        UNMARK(i);
        num_marked_edges--;
        //marked_edges_bitvector &= ~BIT(i);        

    }
    
}

/******************************************************************************/

int contains_valid_2_edge_colouring() {

    //Compute bitvectors with the neighbourhood of each vertex
    int i, j;
    for(i = 0; i < nv; i++) {
        neighbourhood[i] = 0;
        for(j = 0; j < adj[i]; j++)
            neighbourhood[i] |= BIT(current_graph[i][j]);
    }    
    
    if((nv % 2) != 0) {
        fprintf(stderr, "Error: graph must have an even number of vertices!\n");
        exit(1);
    }

    num_edges = 0;
    for(i = 0; i < nv; i++) {
        for(j = 0; j < adj[i]; j++)
            if(i < current_graph[i][j]) {
                edgelist[num_edges][0] = i;
                edgelist[num_edges][1] = current_graph[i][j];
                
                edgelabel[i][current_graph[i][j]] = num_edges;
                edgelabel[current_graph[i][j]][i] = num_edges;
                
                num_edges++;
            }
    }
    
    if(num_edges != (3*nv / 2)) {
        fprintf(stderr, "Error: invalid number of edges: %d\n", num_edges);
        exit(1);
    }
    
    
    //Hack no isomorphisms!
    if(!test_thomassen_conjecture) {
        if((num_edges % 2) != 0) {
            fprintf(stderr, "Error: num edges mod 2 must be 0! (num edges: %d)\n", num_edges);
            exit(1);
        }
    }
    
    
    //Storing edges in unsigned long long int bitvectors!
    if(num_edges > 64) {
        fprintf(stderr, "Error: can only process graphs with at most 64 edges!\n");
        exit(1);
    }
    
    //Marked edges are coloured black, the unmarked edges are coloured white
    
    //In the beginning the entire graph is unmarked
    
    for(i = 0; i < nv; i++)
        marked_adj[i] = 0;
    
    for(i = 0; i < nv; i++) {
        for(j = 0; j < adj[i]; j++)
            unmarked_graph[i][j] = current_graph[i][j];
        
        unmarked_adj[i] = adj[i];
    }
    
    RESETMARKS;
    num_marked_edges = 0;
    //marked_edges_bitvector = 0;
    
    //Edge 0 must be coloured black (to avoid equivalent colourings)
    MARK(0);
    num_marked_edges++;
    //marked_edges_bitvector |= BIT(0);
    
    //fprintf(stderr, "org graph:\n");
    //printgraph(current_graph, adj);
    
    //fprintf(stderr, "adding/removing edge %d %d\n", edgelist[0][0], edgelist[0][1]);
    
    add_edge_given_graph(marked_graph, marked_adj, edgelist[0][0], edgelist[0][1]);
    remove_edge_from_neighbourhood(unmarked_graph, unmarked_adj, edgelist[0][0], edgelist[0][1]);
    remove_edge_from_neighbourhood(unmarked_graph, unmarked_adj, edgelist[0][1], edgelist[0][0]);
    
    
/*
    fprintf(stderr, "marked graph:\n");
    printgraph(marked_graph, marked_adj);
    
    fprintf(stderr, "unmarked graph:\n");
    printgraph(unmarked_graph, unmarked_adj);        
*/
    
    
    valid_colouring_found = 0;
    
    num_2edge_colourings_generated = 0;
    
    colour_edges_in_all_possible_ways_recursion(1);
    
    //fprintf(stderr, "Number of edge-bisections generated: %d\n", num_2edge_colourings_generated);
    
    return valid_colouring_found;

}

/******************************************************************************/

static void
perform_wordsize_checks() {
    /* Checks to test if the WORDSIZE and sizeof setwords is valid */
    if(WORDSIZE != 32 && WORDSIZE != 64) {
        fprintf(stderr, "Error: invalid value for wordsize: %d\n", WORDSIZE);
        fprintf(stderr, "Valid values are 32 and 64\n");
        exit(1);
    }

    if(MAXN != WORDSIZE) {
        fprintf(stderr, "Error: MAXN is not equal to WORDSIZE: %d vs %d\n", MAXN, WORDSIZE);
        fprintf(stderr, "Please compile with option -DMAXN=WORDSIZE\n");
        exit(1);
    }
    
    if(MAXM != 1) {
        fprintf(stderr, "Error: expected MAXM 1 instead of %d\n", MAXM);
        exit(1);
    }

    if(WORDSIZE == 32) {
        fprintf(stderr, "32 bit mode\n");
        if(sizeof (unsigned int) < 4) {
            fprintf(stderr, "Error: unsigned ints should be at least 32 bit -- sorry.\n");
            exit(1);
        }
        if(sizeof (setword) < 4) {
            fprintf(stderr, "Error: this version relies on 32 bit setwords -- sorry.\n");
            exit(1);
        }
        if(sizeof (setword) != 4) {
            fprintf(stderr, "Error: expected 32 bit setwords\n");
            exit(1);
        }
    } else {
        fprintf(stderr, "64 bit mode\n");
        if(sizeof (setword) != 8) {
            fprintf(stderr, "Error: this version relies on 64 bit setwords -- sorry.\n");
            exit(1);
        }
        if(sizeof (unsigned long long int) != 8) {
            fprintf(stderr, "Error: this version relies on 64 bit unsigned long long ints -- sorry.\n");
            exit(1);
        }
    }
}

/******************************************************************************/

static void
print_help(char * argv0) {
    fprintf(stderr, "Conjecture_Wormald version %s\n", VERSION);
    fprintf(stderr, "Usage: %s [options]\n", argv0);
    fprintf(stderr, "\n");
    fprintf(stderr, "Valid options are:\n");
    fprintf(stderr, "  strong: Search for Strong Wormald Colourings.\n");
    fprintf(stderr, "  thomassen: Test if the graphs can be decomposed in two 4-linear forests.\n");
}

/******************************************************************************/

/*
 *
 */
int main(int argc, char** argv) {
    
    perform_wordsize_checks();
    nauty_check(WORDSIZE, MAXM, MAXN, NAUTYVERSIONID);        
    

    if(sizeof (unsigned long long int) != 8) {
        fprintf(stderr, "Error: this version relies on 64 bit unsigned long long ints -- sorry.\n");
        exit(1);
    }    
    
    if(argc == 2 && argv[1][0] == '-' && argv[1][1] == 'h') {
        print_help(argv[0]);
        exit(1);
    }    

    int i;
    for(i = 1; i < argc; i++) {
        switch(argv[i][0]) {
            case 's':
            {
                if(strcmp(argv[i], "strong") == 0) {
                    test_strong_conjecture = 1;
                }
                break;
            }
            case 't':
            {
                if(strcmp(argv[i], "thomassen") == 0) {
                    test_thomassen_conjecture = 1;
                    fprintf(stderr, "Warning: this is a hack which requires that the paths have length at most 4 instead of 5!\n");
                }
                break;
            }            
            default:
            {
                fprintf(stderr, "Error: invalid option: %s\n", argv[i]);
                fprintf(stderr, "Execute '%s -h' for a list of possible options.\n", argv[0]);
                exit(1);
            }
        }

    }
    
    if(test_thomassen_conjecture)
        fprintf(stderr, "Warning: hack without isomorphism testing!\n");
    
    if(test_thomassen_conjecture && test_strong_conjecture) {
        fprintf(stderr, "Error: cannot test strong and Thomassen conjecture at the same time!\n");
        exit(1);
    }
    
    if(test_strong_conjecture)
        fprintf(stderr, "Info: testing strong version of Wormald's conjecture (paths length at least 2)\n");
    else if(test_thomassen_conjecture) {
        fprintf(stderr, "Info: testing Thomassen version of Wormald's conjecture (paths length at most 5)\n");
        fprintf(stderr, "Warning: only testing colourings for which num red edges equals num blue edges!\n");
    } else
        fprintf(stderr, "Info: testing original version of Wormald's conjecture\n");
    
    
    //For the timing
    struct tms TMS;

    int codelength = 0;
    unsigned char code[MAXN * (MAXN - 1) / 2 + MAXN];
    unsigned long long int number_of_graphs_read = 0;
    unsigned long long int number_of_graphs_written = 0;
    
    int nuller;
    for(; fread(code, sizeof (unsigned char), 1, stdin);) {
        nv = code[0];
        if(nv > MAXN) {
            fprintf(stderr, "Number of vertices is too big (limit is %d) \n", MAXN);
            exit(0);
        }
        if(code[0] == 0)
            nuller = 1;
        else
            nuller = 0;
        codelength = 1;
        while(nuller < nv - 1) {
            code[codelength] = getc(stdin);
            if(code[codelength] == 0) nuller++;
            codelength++;
        }

        number_of_graphs_read++;
        decode_multicode(code, codelength);
        
        //fprintf(stderr, "Processing graph %llu\n", number_of_graphs_read);
        //printgraph(graph);
        
        if(!contains_valid_2_edge_colouring()) {
            if(fwrite(code, sizeof (unsigned char), codelength, stdout) < codelength) {
                fprintf(stderr, "Error: couldn't write code!\n");
                exit(1);
            }
            number_of_graphs_written++;
        }
    }
    
    times(&TMS);
    unsigned int savetime = (unsigned int) TMS.tms_utime;
    

    fprintf(stderr, "Read %lld graphs\n", number_of_graphs_read);

    if(test_strong_conjecture)
        fprintf(stderr, "Written %lld graphs which are a counterexample to the strong version of Wormald's conjecture\n", number_of_graphs_written);
    else
        fprintf(stderr, "Written %lld graphs which are a counterexample to Wormald's conjecture\n", number_of_graphs_written);
    fprintf(stderr, "CPU time filter: %.1f seconds.\n", (double) savetime / (double) CLK_TCK);

    return (EXIT_SUCCESS);
}